#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <velib/base/types.h>
#include <velib/base/ve_string.h>
#if CFG_INIT_CANBUS
#include <velib/canhw/canhw.h>
#include <velib/canhw/canhw_driver.h>
#include <velib/canhw/canhw_console.h>
#endif
#include <velib/mk2/hal.h>
#include <velib/platform/task.h>
#include <velib/platform/console.h>
#include <velib/platform/plt.h>
#if CFG_INIT_DEFAULT_DBUS
#include <velib/types/ve_dbus_item.h>
#endif
#include <velib/types/ve_stamp.h>
#include <velib/utils/ve_todo.h>
#include <velib/utils/ve_getopt.h>
#include <velib/utils/ve_logger.h>

#include "platform_init.h"

static const char *program;
static struct option *pltOptions;
static int verbose = 1;
static int dump;
static char *opts;
static int showBanner;

#if CFG_INIT_CANBUS
static VeCanGateway *gw;
static char const* canInterface;
static int silent = veFalse;
#endif

#if CFG_INIT_SERIAL
# ifdef _WIN32
static char *serialDevice = "COM1";
# else
static char *serialDevice = "/dev/ttyS0";
# endif
#endif

static int logBefore;
static int logAfter;

#if !CFG_WITH_LIBEVENT
void pltWatchFileDescriptor(int fd)
{
	VE_UNUSED(fd);
}
#endif

/**
 * Console options shared by all simple tools.
 */
struct option pltDefaultOptions[] =
{
	/* These options set a flag. */
	{"dump",	no_argument,		0,				'd'},
	{"help",	no_argument,		0,				'h'},
	{"verbose",	no_argument,		0,				'v'},
	{"log-before ",	required_argument,	&logBefore,	0},
	{"log-after",	required_argument,	&logAfter,	0},

#if CFG_INIT_CANBUS
	{"canbus",	required_argument,	0,				'c'},
	{"silent",	no_argument,		&silent,		veTrue},
#endif

#if CFG_INIT_SERIAL
	{"serial",	required_argument,	0,				's'},
#endif

#if CFG_WITH_VERSION
	{"version",	no_argument,		0,				'V'},
#endif

#if CFG_WITH_VERSION || defined(CFG_BOOT_BANNER)
	{"banner",	no_argument,		&showBanner,	veTrue},
#endif

#if CFG_INIT_DEFAULT_DBUS
	{"dbus",	required_argument,	0,				1},
#endif

	{0,			0,					0,				0}
};

void pltOptionsUsage(void)
{
	printf("COMMON OPTIONS:\n");
	printf("\n");

#if CFG_INIT_CANBUS
	printf("  -c, --canbus\n");
	printf("   Specifies the CAN-bus interface to use, formatted as [driver]:[channel].\n");
	printf("   kvaser:1 will for example select the first Kvaser channel.\n");
	printf("   -c ? will list the available gateways.\n");
	printf("\n");
#endif

#if CFG_WITH_VERSION
	printf("  --banner\n");
	printf("   Print the version during startup.\n");
	printf("\n");
#endif

#if CFG_INIT_DEFAULT_DBUS
	printf("  --dbus address");
	printf("   Use specified address to connect to the dbus, e.g.\n");
	printf("     - session\n");
	printf("     - system\n");
	printf("     - unix:abstract=/tmp/dbus-O4H8uXQdeA\n");
	printf("\n");
#endif

#if CFG_INIT_CANBUS
	printf("  -d, --dump\n");
	printf("   Dump the contents of the CAN bus frames. Specifying multiple times increases verbosity.\n");
	printf("\n");
#endif

	printf("  -h, --help\n");
	printf("   Print this help.\n");
	printf("\n");

	printf("  --log-before <count>\n");
	printf("   When set to non zero <count> number of log lines, which are\n");
	printf("   not enabled themselves, are nevertheless shown before an\n");
	printf("   log which is enabled. For example an error message will be\n");
	printf("   preceeded by <count> warning and info messages which occured\n");
	printf("   before the error. Combined with -v also warning messages will\n");
	printf("   be preceeded by info messages\n");
	printf("\n");

	printf("  --log-after <count>\n");
	printf("   Like -log-before, but temporarily enables tracing of disabled messages\n");
	printf("   after a warning or error occured.\n");
	printf("\n");

#if CFG_INIT_SERIAL
	printf("  -s, --serial\n");
	printf("    the serial port to use (default %s).\n", serialDevice);
	printf("\n");
#endif

#if CFG_INIT_CANBUS
	printf("  --silent\n");
	printf("   Only listen, no messages or acks are sent\n");
	printf("\n");
#endif

	printf("  -v --verbose\n");
	printf("   Enable log messages. Specifying multiple times increases verbosity.\n");
	printf("\n");

#if CFG_WITH_VERSION
	printf("  -V --version\n");
	printf("   Print the version number.\n");
	printf("\n");
#endif
}

static void printVersion(char const *program)
{
#if CFG_WITH_VERSION
	printf("%s %s\n", program, pltProgramVersion());
#else
	printf("%s starting\n", program);
#endif

#ifdef CFG_COMPILE_STRING
	/*
	 * Addional comment, e.g. compile options etc. Intended for the
	 * build args etc. or anyother relavant build information.
	 * Pass e.g. for gcc with -DCCFG_COMPILE_STRING=\"example\".
	 */
	printf("%s\n", CFG_COMPILE_STRING);
#endif

	fflush(stdout);
}

/**
 * Parse the common option and pass unknown options to the application
 */
void pltParseArgs(int argc, char *argv[])
{
	int c;
	int n;
	int nOptions = ARRAY_LENGTH(pltDefaultOptions) - 1;
	char* ptr;
	char const *p;

	program = argv[0];
	p = program + strlen(program);
	while (p != program) {
		if (*p == '\\' || *p ==  '/') {
			program = p + 1;
			break;
		}
		p--;
	}

	veStampInit();
	logI("task", "start");

	n = 0;
	while (consoleOptions[n].name != NULL)
	{
		nOptions++;
		n++;
	}

	pltOptions = malloc((nOptions +1) * sizeof(struct option));
	if (pltOptions  == NULL)
		exit(-1);

	memcpy(pltOptions, consoleOptions, n * sizeof(struct option));
	memcpy(&pltOptions[n], pltDefaultOptions, sizeof(pltDefaultOptions));

	opts = malloc(3*nOptions+1);
	if (opts == NULL)
		pltExit(-1);

	// create the short options string.
	ptr = opts;
	for (n=0; n<nOptions; n++)
	{
		if (pltOptions[n].flag != 0)
			continue;

		*ptr++ = pltOptions[n].val;

		if (pltOptions[n].has_arg == no_argument)
			continue;

		*ptr++ = ':';
		if (pltOptions[n].has_arg == required_argument)
			continue;

		*ptr++ = ':';
	}
	*ptr = 0;

	for (;;)
	{
		/* getopt_long stores the option index here. */
		int option_index = 0;

		c = getopt_long(argc, argv, opts, pltOptions, &option_index);

		/* Detect the end of the options. */
		if (c == -1)
			break;

		switch (c)
		{
		case 0:
			if (pltOptions[option_index].flag == &logBefore)
				logBefore = atoi(optarg);

			if (pltOptions[option_index].flag == &logAfter)
				logAfter = atoi(optarg);

			break;

#if CFG_INIT_DEFAULT_DBUS
		case 1:
			veDbusSetDefaultConnectString(optarg);
			break;
#endif

#if CFG_INIT_CANBUS
		case 'c':
			canInterface = optarg;
			break;
#endif

		case 'd':
			dump++;
			break;

		case 'h':
			consoleUsage(argv[0]);
			pltExit(0);
			break;

#if CFG_INIT_SERIAL
		case 's':
			serialDevice = optarg;
			break;
#endif

		case 'v':
			verbose++;
			break;

#if CFG_WITH_VERSION
		case 'V':
			printVersion(program);
			pltExit(0);
			break;
#endif

		default:
			if (consoleOption(c))
				continue;

		case '?':
			/* getopt_long already printed an error message. */
			pltExit(-1);
		}
	}

	veLogLevel(verbose);
	veLogBackBufferLines(logBefore);
	veLogLinesAfterTrigger(logAfter);

	if (showBanner)
		printVersion(program);

	/* pass remaing argument to the console application */
	if (!consoleArgs(argc - optind, &argv[optind]))
		pltExit(-1);

	free(pltOptions);
}

#if CFG_INIT_CANBUS

static veBool canRxEventCb(struct VeCanGatewayS* gw)
{
	VE_UNUSED(gw);

	veTodo();
	return veTrue;
}

/**
 * Open the canbus as specified by the console arguments.
 * Currently only 1 canbus interface is supported.
 */
void pltInitCanBus(void)
{
	VeCanRxEventType rxEvType;
	VeCanRxEventData rxEvData;

	// start CAN bus
	veCanInit();

	/* If the gateway is passed as an argument used that */
	if (canInterface && strcmp(canInterface, "?") != 0) {
		/*
		 * Problems have been reported that it might take a long time
		 * loading and enumerating all drivers. (typically serial drivers
		 * in combination with fake uarts can cause long delays. So if a
		 * driver is explicitly given, only load and enumerate that one.
		 */
		char *p;
		char *module = strdup(canInterface);
		if ( (p = strchr(module, ':')) != NULL)
			*p = 0;
#if !CFG_CANHW_DRIVER_STATIC
		veCanDrvLoad(module);
#endif
		free(module);
		veCanDrvEnumerate();
		gw = veCanGwGet(canInterface);
		if (!gw)
			printf("Specified CAN-bus gateway '%s' not found!\n", canInterface);

	} else {

#if !CFG_CANHW_DRIVER_STATIC
		/* if no specific driver is given, just load them all  */
		veCanDrvDynamicInit();
#endif
		/* and enumerate all drivers */
		veCanDrvEnumerate();
		logI("task", "enumerate done");

		gw = veCanGwFindDefault();
		if (!gw) {
			printf("None or more than one physical gateway found.\n");
			printf("Use -c to specify which gateway should be used.\n");
		}
	}

	if (!gw || (canInterface && strcmp(canInterface, "?") == 0))
	{
		printf("The following gateways were found:\n");
		printf("--------------------------------------------------------------\n");
		veCanDrvPrintGwList();
		printf("--------------------------------------------------------------\n");
		pltExit(veCanDrvGetGwCount() == 0 ? 9 : 15);
	} else {
		char buf[100];
		if (veCanGwDesc(gw, buf, sizeof(buf)) < sizeof(buf))
			logI("", "Using %s", buf);
	}

	if (!veCanOpen())
	{
		logE("CAN", "canOpen failed");
		pltExit(10);
	}

	if (veCanSetBitRate(250) == veFalse)
	{
		logE("CAN", "set bitrate failed");
		pltExit(11);
	}

	veCanShowTrace(dump);

	/* note: must be after set bitrate */
	if (silent) {
		//if (!veGeSetSilent(veTrue)) {
		/* note: broken at the moment */
			logE("CAN", "disable acking failed");
			pltExit(14);
		//}
	}

	rxEvType = veCanGwRxEventType(gw, &rxEvData);
	switch (rxEvType)
	{
	case VE_CAN_RX_NONE:
		break;
	case VE_CAN_RX_EV_ASYNC_CALLBACK:
		veCanGwSetRxEventCb(gw, canRxEventCb);
		break;
	case VE_CAN_RX_EV_FD:
		pltWatchFileDescriptor(rxEvData.fd);
		break;
	}

	if (!veCanBusOn())
	{
		logE("CAN", "can bus on failed");
		pltExit(12);
	}
}
#else
void pltInitCanBus(void)
{
}
#endif

#if CFG_INIT_SERIAL_MK2
void pltMk2Init(void)
{
	logI("mk2", "opening port %s", serialDevice);
	if (!mk2Open(serialDevice)) {
		logE("init", "cannot open '%s'", serialDevice);
		pltExit(1);
	}
}
#else
void pltMk2Init(void)
{
}
#endif

/** common console startup */
void consoleInit(int argc, char *argv[])
{
	veStampInit();
	logI("task", "start");

	pltParseArgs(argc, argv);

#if CFG_INIT_CTX
	/*
	 * note: this will "disable" interrupts during init, to behave more
	 * or less how sillicon will get out of reset. Which are stubs on
	 * bare targets and mutexes on an threaded OS. It is left up to the
	 * application when to "enable" interrupts with pltInterruptEnable();
	 * but typically that is add the end of taskInit.
	 */
	pltInitCtx();
#endif

#if CFG_TASK_EARLY_INIT
	taskEarlyInit();
#endif

	pltInitCanBus();

	pltMk2Init();

	taskInit();
}

void pltExit(int ret)
{
	VeStamp time;

	veStampGetAbs(&time);
	logI("task", "done in %03d.%03d", (un32) (time.us / 1000000), (un32) ((time.us / 1000) % 1000));

	exit(ret);
}

#if CFG_INIT_SERIAL
char const *pltGetSerialDevice(void)
{
	return serialDevice;
}
#endif

char const *pltProgramName(void)
{
	return program;
}
